"""
  CS 218, UC Riverside
  Fast Fourier Transform
  adapted from http://numericalrecipes.blogspot.com/2009/04/cooley-turkey-fft-algorithm.html
  
"""

from __future__ import division
import math

def dft(x, inverse = False) :
    N = len(x)
    if not inverse:
	inv = -1
    else:
	inv = 1
    X =[0] * N
    for k in range(N) :
        for n in range(N) :
            X[k] += x[n] * math.e**(inv * 2j * math.pi * k * n / N)
        if inverse :
            X[k] /= N
    return X

def fft(x, inverse = False) :
    N = len(x)
    if not inverse:
        w_n = math.e ** (-2j * math.pi / N) # root of unity
    else:
        w_n = math.e ** (2j * math.pi / N) # root of unity
    if N % 2 : # if N is odd
        return dft(x, inverse)
    x_e = x[::2] # even coefficients
    x_o = x[1::2] # odd coefficients
    X_e = fft(x_e, inverse) # DFT of even coefficients
    X_o = fft(x_o, inverse) # DFT of odd coefficients
    X = []
    M = N // 2
    w = 1
    for k in range(M) : # first half
        X += [X_e[k] + X_o[k] * w]
	w *= w_n
    w = 1
    for k in range(M,N) : # second half
        X += [X_e[k-M] - X_o[k-M] * w]
        w *= w_n
    if inverse :
        X = [j/2 for j in X]
    return X

#a = [0, 1, 2, 3, 4, 5, 6, 7]
#print 'input vector a=', a
#print ' --------- DIRECT -----------'
#c = fft(a)
#print 'fast DFT(a)=', c
#print
#d = dft(a)
#print 'slow DFT(a)=', d
#print ' --------- INVERSE -----------'
#print 'fast DFT^-1(DFT(a))=', fft(c, True)
#print
#print 'slow DFT^-1(DFT(a))=', dft(d, True)

print ' --------------------------------------------'
print ' --------- POLYNOMIAL MULTIPLICATION --------'
print ' --------------------------------------------'
a = [2, 1, 1, 3, 0, 0, 0, 0] # polynomial a
b = [1, 2, 3, 1, 0, 0, 0, 0] # polynomial b
print 'a(x) = [a_0, a_1, a_2, a_3, a_4, a_5, a_6, a_7] =', a
print 'b(x) = [b_0, b_1, b_2, b_3, b_4, b_5, b_6, b_7] =', b

c = [0, 0, 0, 0, 0, 0, 0, 0] # product polynomial 
for i in range(len(a)): # quadratic time convolution 
    for j in range(i+1):
        c[i] += a[j]*b[i-j]
print 'naive c(x)=', c

aT = fft(a) # O(n log n) time
bT = fft(b) # O(n log n) time
cT = [0, 0, 0, 0, 0, 0, 0, 0]
for i in range(len(aT)): # pointwise multiply (linear-time)
    cT[i] = aT[i]*bT[i]
print 'FFT c(x)=', fft(cT, True) # inverse FFT O(n log n)

