"""
  CS 141/218, UC Riverside, Stefano Lonardi

  An implementation of naive and Strassen's algorithm for matrix multiplication

"""

import random
 
def zero(m,n):
    """
    Create zero matrix
    """
    new_matrix = [[0 for row in range(n)] for col in range(m)]
    return new_matrix
 
def rand(m,n):
    """ 
    Create random matrix
    """
    new_matrix = [[int(random.random()*10) for row in range(n)] for col in range(m)]
    return new_matrix
 
def show(matrix):
    """
    Print out matrix
    """
    for col in matrix:
        print col 

def partition(matrix):
    """
    partition a squared even-sided matrix in four equal pieces
    """
    n = len(matrix)
    m = len(matrix[0])
    if n != m:
        print 'Matrix has to be squared'
    else:
        middle = n / 2
        first_half = matrix[:middle]
        A = []
        for col in first_half:
            A += [col[:middle]]
        B = []
        for col in first_half:
            B += [col[middle:]]
        second_half = matrix[middle:]
        C = []
        for col in second_half:
            C += [col[:middle]]
        D = []
        for col in second_half:
            D += [col[middle:]]
        return (A,B,C,D)

def recompose(m1,m2,m3,m4):
    """
    returns a matrix composed of
    |m1|m2|
    |m3|m4|
    """
    result = []
    for i in range(len(m1)):
        result += [m1[i]+m2[i]]
    for i in range(len(m3)):
        result += [m3[i]+m4[i]]
    return result

def strassen(matrix1,matrix2):
    """
    compute the product of squared matrices matrix1 and matrix2 using
    Strassen's algorithm
    """
    n = len(matrix1)     
    m = len(matrix1[0])
    p = len(matrix2)
    q = len(matrix2[0])
    if n != m and m != p and p != q: 
        # Check matrix dimensions
        print 'Matrices must be squared and same dimentions to multiply!'
    else:
        #print n,m,p,q
        if n <= 2:
            return mult(matrix1, matrix2)
        else:
            (A,B,C,D) = partition(matrix1)
            (E,F,G,H) = partition(matrix2)
            M1 = strassen(add(A,C),add(E,F))
            M2 = strassen(add(B,D),add(G,H))
            M3 = strassen(sub(A,D),add(E,H))
            M4 = strassen(A,sub(F,H))
            M5 = strassen(add(C,D),E)
            M6 = strassen(add(A,B),H)
            M7 = strassen(D,sub(G,E))
            I = sub(sub(add(M2,M3),M6),M7)
            J = add(M4,M6)
            K = add(M5,M7)
            L = sub(sub(sub(M1,M3),M4),M5)
            return recompose(I,J,K,L)

def mult(matrix1,matrix2):
    """
    Naive matrix multiplication
    """
    if len(matrix1[0]) != len(matrix2):
        # Check matrix dimensions
        print 'Matrices must be m*n and n*p to multiply!'
    else:
        # Multiply if correct dimensions
        new_matrix = zero(len(matrix1),len(matrix2[0]))
        for i in range(len(matrix1)):
            for j in range(len(matrix2[0])):
                for k in range(len(matrix2)):
                    new_matrix[i][j] += matrix1[i][k]*matrix2[k][j]
        return new_matrix

def sub(matrix1,matrix2):
    """
    Matrix subtraction
    """
    if len(matrix1[0]) != len(matrix2[0]) and len(matrix1) != len(matrix2):
        # Check matrix dimensions
        print 'Matrices must have same dimention to sum!'
    else:
        new_matrix = zero(len(matrix1),len(matrix1[0]))
        for i in range(len(matrix1)):
            for j in range(len(matrix1[0])):
                new_matrix[i][j] = matrix1[i][j]-matrix2[i][j]
        return new_matrix

def add(matrix1,matrix2):
    """
    Matrix addition
    """
    if len(matrix1[0]) != len(matrix2[0]) and len(matrix1) != len(matrix2):
        # Check matrix dimensions
        print 'Matrices must have same dimention to sum!'
    else:
        new_matrix = zero(len(matrix1),len(matrix1[0]))
        for i in range(len(matrix1)):
            for j in range(len(matrix1[0])):
                new_matrix[i][j] = matrix1[i][j]+matrix2[i][j]
        return new_matrix

if __name__ == "__main__":
    a = rand(8,8)
    b = rand(8,8)
    show(a)
    print
    show(b)
    print
    print 'Naive multiplication'
    c = mult(a,b)
    show(c)
    print
    print 'Strassen multiplication'
    d = strassen(a,b)
    show(d)
